from homealone import *
from homealone.rest.restConfig import *
from homealone.rest.restService import *
from homealone.rest.restInterface import *
import json
import threading
import socket
import time

# Proxy for the resources exposed by one or more REST services

# Usecases:
#     proxy detects new service
#     service states updated
#     service resources updated
#     rest i/o error
#     notify message timeout
#     service restarts

# set default service port if not specified
def setServicePorts(serviceList):
    newServiceList = []
    for service in serviceList:
        if len(service.split(".")) < 2:
            service = "services."+service
        newServiceList.append(service)
    return newServiceList

def parseServiceData(data, addr):
    try:
        serviceData = json.loads(data.decode("utf-8"))
        debug('debugRestProxyData', "data", serviceData)
        try:
            serviceResources = serviceData["resources"]
        except KeyError:
            serviceResources = None
        try:
            serviceStates = serviceData["states"]
        except:
            serviceStates = None
        serviceData = serviceData["service"]
        serviceName = "services."+serviceData["name"]
        serviceAddr = addr[0]+":"+str(serviceData["port"])
        try:
            stateTimeStamp = serviceData["statetimestamp"]
            resourceTimeStamp = serviceData["resourcetimestamp"]
        except KeyError:
            stateTimeStamp = serviceData["timestamp"]
            resourceTimeStamp = serviceData["timestamp"]
        try:
            version = serviceData["version"]
        except KeyError:
            version = 0
        serviceLabel = serviceData["label"]
        serviceSeq = serviceData["seq"]
        return (serviceName, serviceAddr, serviceLabel, version, serviceSeq, stateTimeStamp, resourceTimeStamp, serviceStates, serviceResources)
    except Exception as ex:
        logException("parseServiceData", ex)
        return ("", "", "", 0, 0, 0, 0, {}, [])

# Autodiscover services and resources
# Detect changes in resource configuration on each service
# Remove resources on services that don't respond

class RestProxy(LogThread):
    def __init__(self, name, resources, watch=[], ignore=[], event=None, cache=True):
        debug('debugRestProxy', name, "starting", name)
        LogThread.__init__(self, target=self.restProxyThread)
        self.name = name
        self.services = {}                      # cached services
        self.resources = resources              # resource cache
        self.event = event
        self.cache = cache
        self.cacheTime = 0                      # time of the last update to the cache
        self.watch = setServicePorts(watch)     # services to watch for
        self.ignore = setServicePorts(ignore)   # services to ignore
        # self.ignore.append("services."+socket.gethostname()+":"+str(restServicePort))   # always ignore services on this host
        debug('debugRestProxy', name, "watching", self.watch)    # watch == [] means watch all services
        debug('debugRestProxy', name, "ignoring", self.ignore)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind((multicastAddr, restNotifyPort))

    def restProxyThread(self):
        debug('debugThread', self.name, "started")
        while True:
            # wait for a notification message from a service
            (data, addr) = self.socket.recvfrom(32768)   # FIXME - need to handle arbitrarily large data
            debug('debugRestMessage', self.name, "notification data", data)
            # parse the message
            (serviceName, serviceAddr, serviceLabel, version, serviceSeq, stateTimeStamp, resourceTimeStamp, serviceStates, serviceResources) = \
                parseServiceData(data, addr)
            if serviceName == "":   # message couldn't be parsed
                continue
            # rename it if there is an alias
            if serviceName in list(self.resources.aliases.keys()):
                newServiceName = self.resources.aliases[serviceName]["name"]
                newServiceLabel = self.resources.aliases[serviceName]["label"]
                debug('debugRestProxy', self.name, "renaming", serviceName, "to", newServiceName)
                serviceName = newServiceName
                serviceLabel = newServiceLabel
            # determine if this service should be processed based on watch and ignore lists
            if ((self.watch != []) and (serviceName  in self.watch)) or ((self.watch == []) and (serviceName not in self.ignore)):
                debug('debugRestProxy', self.name, "processing", serviceName, serviceAddr, stateTimeStamp, resourceTimeStamp)
                if serviceName not in list(self.services.keys()):
                    # this is one not seen before, create a new service proxy
                    debug('debugRestProxyAdd', self.name, "adding", serviceName, serviceAddr, version, stateTimeStamp, resourceTimeStamp)
                    self.services[serviceName] = RestService(serviceName,
                                                            RestInterface(serviceName+"Interface",
                                                                            serviceAddr=serviceAddr,
                                                                            event=self.event,
                                                                            cache=self.cache),
                                                            version=version,
                                                            proxy=self,
                                                            label=serviceLabel,
                                                            group="Services")
                    service = self.services[serviceName]
                    service.enable()
                else:   # service is already in the cache
                    service = self.services[serviceName]
                    service.cancelTimer("message received")
                    if not service.enabled:     # the service was previously disabled but it is broadcasting again
                        # re-enable it
                        debug('debugRestProxyDisable', self.name, "reenabling", serviceName, serviceAddr, version, stateTimeStamp, resourceTimeStamp)
                        # update the resource cache
                        # self.addResources(service)
                        service.interface.serviceAddr = serviceAddr # update the ipAddr:port in case it changed
                        service.enable()
                # load the resources or states in a separate thread if there was a change
                if (resourceTimeStamp > service.resourceTimeStamp) or serviceResources or \
                   (stateTimeStamp > service.stateTimeStamp) or serviceStates:
                    if not service.updating:    # prevent multiple updates at the same time
                        service.updating = True
                        startThread(serviceName+"-update", self.loadService, args=(service, resourceTimeStamp, serviceResources,
                                                                                stateTimeStamp, serviceStates,))
                # start the message timer
                service.startTimer()
                service.logSeq(serviceSeq)
            else:
                debug('debugRestProxy', self.name, "ignoring", serviceName, serviceAddr, stateTimeStamp, resourceTimeStamp)
        debug('debugThread', self.name, "terminated")

    # optionally load the resources for a service and add them to the proxy cache
    # optionally load the states of the service resources
    def loadService(self, service, resourceTimeStamp, serviceResources, stateTimeStamp, serviceStates):
        debug('debugThread', threading.currentThread().name, "started")
        try:
            if (resourceTimeStamp > service.resourceTimeStamp) or serviceResources:
                debug('debugRestProxyUpdate', self.name, "updating resources", service.name, resourceTimeStamp)
                service.load(serviceResources)
                service.resourceTimeStamp = resourceTimeStamp
                self.addResources(service)
            if (stateTimeStamp > service.stateTimeStamp) or serviceStates:
                debug('debugRestProxyStates', self.name, "updating states", service.name, stateTimeStamp)
                if not serviceStates:
                    # if states not provided, get them from the service
                    serviceStates = service.interface.getStates()
                else:
                    service.interface.setStates(serviceStates)     # load the interface cache
                self.resources.setStates(serviceStates)        # update the resource collection cache
                service.stateTimeStamp = stateTimeStamp
                self.resources.notify()
        except Exception as ex:
            logException(self.name+" loadService", ex)
        service.updating = False
        debug('debugThread', threading.currentThread().name, "terminated")

    # add the resource of the service as well as
    # all the resources from the specified service to the cache
    def addResources(self, service):
        debug('debugRestProxy', self.name, "adding resources for service", service.name)
        self.resources.addRes(service, 1)                       # the resource of the service
        self.resources.addRes(service.missedSeqSensor)          # missed messages
        self.resources.addRes(service.missedSeqPctSensor)       # percent of missed messages
        for resource in list(service.resources.values()):       # resources from the service
            self.resources.addRes(resource)
        self.cacheTime = service.resourceTimeStamp # FIXME
        self.event.set()
        debug('debugInterrupt', self.name, "event set")

    # disable all the resources from the specified service from the cache
    # don't delete the resource of the service
    # def disableResources(self, service):
    #     debug('debugRestProxyDisable', self.name, "disabling resources for service", service.name)
    #     for resourceName in list(service.resources.keys()):
    #         try:
    #             self.resources.getRes(resourceName).enabled = False
    #         except KeyError:
    #             debug('debugRestProxyDisable', service.name, "error disabling", resourceName)
    #     self.event.set()
    #     debug('debugInterrupt', self.name, "event set")



from homealone import *
from homealone.resources.extraResources import *
from homealone.rest.restConfig import *
from homealone.schedule import *
import threading

messageTimeout = beaconTimeout

# create a Resource from a serialized dict
def loadResource(classDict, globalDict):
    def parseClass(classDict):
        args = classDict["args"]
        argStr = ""
        for arg in list(args.keys()):
            argStr += arg+"="
            if isinstance(args[arg], dict):     # argument is a class
                argStr += parseClass(args[arg])+", "
            elif isinstance(args[arg], str):    # arg is a string
                argStr += "'"+args[arg]+"', "
            elif not arg:                       # arg is None
                argStr += "None"
            else:                               # arg is numeric or other
                argStr += str(args[arg])+", "
        return classDict["class"]+"("+argStr[:-2]+")"
    localDict = {}
    exec("resource = "+parseClass(classDict), globalDict, localDict)
    return localDict["resource"]

# proxy for a REST service
class RestService(Sensor):
    def __init__(self, name, interface, addr=None, version=0, stateTimeStamp=-1, resourceTimeStamp=-1, proxy=None, type="service", **kwargs):
        Sensor.__init__(self, name, interface, addr=addr, type=type, **kwargs)
        debug('debugRestService', "RestService", name, "created")
        self.version = version
        self.stateTimeStamp = stateTimeStamp      # the last time the states were updated
        self.resourceTimeStamp = resourceTimeStamp      # the last time the resources were updated
        self.resources = Collection(self.name+"/Resources")           # resources on this service
        self.proxy = proxy              # RestProxy that is following this service
        self.enabled = False
        self.messageTimer = None
        self.updating = False
        self.lastSeq = 0                # the last message sequence number received
        self.missedSeq = 0              # count of how many missed messages for this service
        self.missedSeqPct = 0.0         # percentage of missed messages
        try:
            serviceName = name.split(".")[1]
        except IndexError:
            serviceName = name
        self.missedSeqSensor = AttributeSensor(serviceName+"-missedSeq", None, None, self, "missedSeq")
        self.missedSeqPctSensor = AttributeSensor(serviceName+"-missedSeqPct", None, None, self, "missedSeqPct")

    def getState(self, missing=None):
        return normalState(self.enabled)

    def setState(self, state, wait=False):
        if state:
            self.enable()
        else:
            self.disable("set")
        return True

    def __repr__(self):
        return "version: "+str(self.version)+"\n"+ \
               "resource time: "+str(self.resourceTimeStamp)+"\n"+ \
               "state time: "+str(self.stateTimeStamp)+"\n"+ \
               "missed seq: "+str(self.missedSeq)+"\n"+ \
               "---------------"

    def enable(self):
        debug('debugRestService', "RestService", self.name, "enabled")
        for resource in list(self.resources.values()):
            resource.enable()
        self.interface.start()
        self.enabled = True
        self.notify(True)

    def disable(self, reason=""):
        debug('debugRestService', "RestService", self.name, "disabled", reason)
        self.enabled = False
        self.interface.stop()
        self.messageTimer = None
        for resource in list(self.resources.values()):
            resource.disable()
        self.notify(False)

    def logSeq(self, seq):
        debug('debugRestSeq', "RestService", self.name, seq, self.lastSeq, self.missedSeq, self.missedSeqPct)
        if seq == 0:
            self.lastSeq = 0    # reset when the service starts
            self.missedSeqPct = 0.0
        if self.lastSeq != 0:   # ignore the first one after this program starts
            self.missedSeq += seq - self.lastSeq - 1
        if seq > 0:
            self.missedSeqPct = float(self.missedSeq) / float(seq)
        self.lastSeq = seq
        self.missedSeqSensor.notify()
        self.missedSeqPctSensor.notify()

    # define a timer to disable the service if the message timer times out
    # can't use a socket timeout because multiple threads are using the same port
    def messageTimeout(self):
        debug('debugMessageTimer', self.name, "timer expired")
        debug('debugRestProxyDisable', self.name, "read message timeout")
        self.disable("timeout")

    # start the message timer
    def startTimer(self):
        if messageTimeout:
            self.messageTimer = threading.Timer(messageTimeout, self.messageTimeout)
            self.messageTimer.start()
            debug('debugMessageTimer', self.name, "timer started", messageTimeout, "seconds")

    # cancel the message timer
    def cancelTimer(self, reason=""):
        if self.messageTimer:
            self.messageTimer.cancel()
            debug('debugMessageTimer', self.name, "timer cancelled", reason)

    # load resources from the specified REST paths
    def load(self, serviceResources):
        debug('debugLoadService', self.name, "load", serviceResources)
        try:
            # self.delResources()
            # self.addResources()
            if not serviceResources or \
                    (isinstance(serviceResources["args"]["resources"], list)):  # if expanded resources not provided, get them
                serviceResources = self.interface.readRest("/resources?expand=true")
            for resource in serviceResources["args"]["resources"]:
                self.loadResource(resource)
        except Exception as ex:
            logException(self.name+" load", ex)

    # instantiate the resource from the specified dictionary
    def loadResource(self, resourceDict):
        debug('debugLoadService', self.name, "loadResource", resourceDict)
        resourceDict["args"]["interface"] = None
        resource = loadResource(resourceDict, globals())
        # replace the resource interface and addr with the REST interface and addr
        resource.interface = self.interface
        resource.addr = resource.name
        resource.interface.addSensor(resource)
        self.resources.addRes(resource)




metricsPrefix = "com.example.ha"
metricsHost = "metrics.example.com"
metricsPort = 2003

import time
import socket
import threading
import json
import subprocess
import os
import socket
from ha import *

def startMetrics(resources, sendMetrics=False, logMetrics=True, backupMetrics=True, purgeMetrics=False, purgeDays=5, logChanged=True):

    def sendMetricsThread():
        debug("debugMetrics", "sendMetrics", "metrics thread started")
        hostname = socket.gethostname()
        lastDay = ""
        while True:
            # wait for a new set of states
            states = resources.getStates(wait=True)
            today = time.strftime("%Y%m%d")

            # log states to a file
            if logMetrics:
                if today != lastDay:    # start with a new baseline every day
                    lastStates = states
                changedStates = diffStates(lastStates, states, deleted=False)
                if changedStates != {}:
                    logFileName = logDir+today+".json"
                    debug("debugMetrics", "sendMetrics", "writing states to", logFileName)
                    with open(logFileName, "a") as logFile:
                        logFile.write(json.dumps([time.time(), (changedStates if logChanged else lastStates)])+"\n")
                lastStates = states

            # send states to the metrics server
            if sendMetrics:
                debug("debugMetrics", "sendMetrics", "opening socket to", metricsHost, metricsPort)
                try:
                    metricsSocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    metricsSocket.connect((metricsHost, metricsPort))
                    debug("debugMetrics", "sendMetrics", "sending", len(states), "metrics")
                    for state in list(states.keys()):
                        if isinstance(states[state], (int, float)):     # only send numeric data
                            if state.split(".")[0] in ["loads", "solar"]:
                                metricsGroup = "."
                            else:
                                metricsGroup = ".ha."
                            msg = metricsPrefix+metricsGroup+state.replace(" ", "_")+" "+str(states[state])+" "+str(int(time.time()))
                            debug("debugMetricsMsg", "sendMetrics", msg)
                            metricsSocket.send(bytes(msg+"\n", "utf-8"))
                        else:
                            debug("debugMetrics", "sendMetrics", "skipping", state, states[state])
                except socket.error as exception:
                    log("sendMetrics", "socket error", str(exception))
                if metricsSocket:
                    debug("debugMetrics", "sendMetrics", "closing socket to", metricsHost)
                    metricsSocket.close()

            # copy to the backup server once per day
            if backupMetrics:
                if today != lastDay:
                    def backupMetricsThread():
                        debug("debugMetrics", "sendMetrics", "backup thread started")
                        try:
                            backupServer = subprocess.check_output("avahi-browse -atp|grep _backup|grep IPv4|cut -d';' -f4", shell=True).decode().split("\n")[0]+".local"
                            debug("debugMetrics", "backupMetrics", "backing up "+logDir+" to", backupServer)
                            pid = subprocess.Popen("rsync -a "+logDir+"* "+backupServer+":/backups/ha/"+hostname+"/", shell=True)
                        except Exception as ex:
                            log("metrics", "exception backing up metrics", str(ex))
                        debug("debugMetrics", "sendMetrics", "metrics thread ended")
                    backupThread = LogThread(name="backupThread", target=backupMetricsThread)
                    backupThread.start()

            # purge metrics that have been backed up
            if purgeMetrics:
                if today != lastDay:
                    backupServer = subprocess.check_output("avahi-browse -atp|grep _backup|grep IPv4|cut -d';' -f4", shell=True).decode().split("\n")[0]+".local"
                    # get list of metrics files that are eligible to be purged
                    debug("debugPurgeMetrics", "purging metrics more than", purgeDays, "days old")
                    for metricsFile in sorted(os.listdir(logDir))[:-purgeDays]:
                        # only purge past files
                        debug("debugPurgeMetrics", "checking", metricsFile)
                        if metricsFile.split(".")[0] < today:
                            try:
                                # get sizes of the file and it's backup
                                fileSize = int(subprocess.check_output("ls -l "+logDir+metricsFile+"|cut -f5 -d' '", shell=True))
                                backupSize = int(subprocess.check_output("ssh "+backupServer+" ls -l /backups/ha/"+hostname+"/"+metricsFile+"|cut -f5 -d' '", shell=True))
                                if backupSize == fileSize:
                                    debug("debugPurgeMetrics", "deleting", metricsFile)
                                    os.remove(logDir+metricsFile)
                            except Exception as ex:
                                log("exception purging metrics file", metricsFile, str(ex))

            if today != lastDay:
                lastDay = today
        debug("debugMetrics", "sendMetrics", "metrics thread ended")

    metricsThread = LogThread(name="metricsThread", target=sendMetricsThread)
    metricsThread.start()
